import json
import inspect
import logging
import traceback
import functools

from zenutils import importutils
from pydantic import BaseModel
from pydantic import ValidationError

from django.http import HttpResponseBase
from django.http import JsonResponse
from django.conf import settings

from .exceptions import BizError
from .exceptions import MethodNotAllowed
from .exceptions import UnsupportedMediaType
from .exceptions import RequestValidationError
from .exceptions import InternalServerError
from .constants import DJANGO_APIS_FUNC_PARAMETERS_KEY
from .constants import DJANGO_APIS_METHODS_KEY
from .constants import DJANGO_APIS_VIEW_FLAG_KEY
from .constants import DJANGO_APIS_APIVIEW_INSTANCE_KEY
from .schemas import SimpleResponse

_logger = logging.getLogger(__name__)


class Apiview(object):
    base_response_data_field = "data"
    base_response_class = SimpleResponse

    def __call__(self, methods="GET"):
        methods = self.get_methods(methods)

        def view(func):
            def inner_view(request, **path_kwargs):
                try:
                    self.request_method_check(request, methods)
                    func_data = self.get_func_data(
                        func,
                        request,
                        path_kwargs,
                    )
                    result = func(**func_data)
                    if isinstance(result, HttpResponseBase):
                        return result
                    else:
                        if isinstance(result, BaseModel):
                            result = result.dict()
                        return self.make_response(result)
                # 请求参数检验错误
                except ValidationError:
                    error = RequestValidationError()
                    return self.make_error_response(
                        error.code,
                        error.message,
                        status_code=error.code,
                    )
                # json payload解析错误
                except UnsupportedMediaType as error:
                    return self.make_error_response(
                        error.code,
                        error.message,
                        status_code=415,
                    )
                # http请求方法检验错误
                except MethodNotAllowed as error:
                    return self.make_error_response(
                        error.code,
                        error.message,
                        status_code=405,
                    )
                # 其它业务逻辑错误
                except BizError as error:
                    return self.make_error_response(
                        error.code,
                        error.message,
                    )
                # 未处理的系统错误
                except Exception:
                    _logger.exception(
                        "InternalServerError: PATH=%s, GET=%s, POST=%s, FILES=%s, META=%s",
                        request.path,
                        request.GET,
                        request.POST,
                        request.FILES,
                        request.META,
                    )
                    error = InternalServerError()
                    return self.make_error_response(
                        error.code,
                        error.message,
                        status_code=error.code,
                    )

            setattr(func, DJANGO_APIS_VIEW_FLAG_KEY, True)
            setattr(func, DJANGO_APIS_METHODS_KEY, methods)
            setattr(func, DJANGO_APIS_APIVIEW_INSTANCE_KEY, self)
            setattr(func, "csrf_exempt", True)
            return functools.wraps(func)(inner_view)

        return view

    def get_methods(self, methods):
        if isinstance(methods, str):
            methods = [x.strip().upper() for x in methods.split(",")]
            methods = list(set(methods))
            methods.sort()
            return methods
        elif isinstance(methods, (list, str, tuple)):
            methods = list(set(methods))
            methods = [x.strip().upper() for x in methods]
            methods.sort()
            return methods
        else:
            _logger.warning(
                """django-apis' apiview get bad methods=%s, change it to the default value ["GET"].""",
                methods,
            )
            traceback.print_stack()
            return ["GET"]

    def request_method_check(self, request, methods):
        if not request.method in methods:
            raise MethodNotAllowed()

    def make_response(self, data):
        return JsonResponse(
            {
                "code": 0,
                "message": "OK",
                "data": data,
            },
            json_dumps_params={
                "ensure_ascii": False,
            },
        )

    def make_error_response(self, code, message, status_code=200):
        return JsonResponse(
            {
                "code": code,
                "message": message,
                "data": None,
            },
            json_dumps_params={
                "ensure_ascii": False,
            },
            status=status_code,
        )

    def get_func_parameters(self, func):
        if hasattr(func, DJANGO_APIS_FUNC_PARAMETERS_KEY):
            return getattr(func, DJANGO_APIS_FUNC_PARAMETERS_KEY)
        func_parameters = inspect.signature(func).parameters
        setattr(func, DJANGO_APIS_FUNC_PARAMETERS_KEY, func_parameters)
        return func_parameters

    def get_func_data(self, func, request, path_kwargs):
        data = {}
        for name, param in self.get_func_parameters(func).items():
            if name == "request":
                data["request"] = request
            elif name == "payload":
                data["payload"] = self.get_func_payload_data(request, param.annotation)
            elif name == "form":
                data["form"] = self.get_func_form_data(request, param.annotation)
            elif name == "query":
                data["query"] = self.get_func_query_data(request, param.annotation)
            else:
                data[name] = path_kwargs.get(name, None)
        return data

    def get_func_query_data(self, request, type):
        query = self.get_clean_query_data(request)
        return self.request_validate(query, type)

    def get_func_form_data(self, request, type):
        form = self.get_clean_form_data(request)
        return self.request_validate(form, type)

    def get_func_payload_data(self, request, type):
        try:
            payload = json.loads(request.body)
        except Exception as error:
            raise UnsupportedMediaType()
        return self.request_validate(payload, type)

    def request_validate(self, data, type):
        if issubclass(type, BaseModel):
            return type.model_validate(data)
        elif callable(type):
            return type(data)
        else:
            return data

    def get_clean_query_data(self, request):
        data = {}
        for key in request.GET.keys():
            value = request.GET.getlist(key)
            if isinstance(value, list) and len(value) == 1:
                value = value[0]
            data[key] = value
        return data

    def get_clean_form_data(self, request):
        """可能为多媒体表单。"""
        data = {}
        for key in request.POST.keys():
            value = request.POST.getlist(key)
            if isinstance(value, list) and len(value) == 1:
                value = value[0]
            data[key] = value
        for key in request.FILES.keys():
            value = request.FILES.getlist(key)
            if isinstance(value, list) and len(value) == 1:
                value = value[0]
            data[key] = value
        return data


apiview = Apiview()



class BizError(RuntimeError):
    def __init__(self, code, message):
        super().__init__(code, message)

    @property
    def code(self):
        return self.args[0]

    @property
    def message(self):
        return self.args[1]


class RequestValidationError(BizError):
    def __init__(self):
        super().__init__(400, "Validation Error")


class MethodNotAllowed(BizError):
    def __init__(self):
        super().__init__(405, "Method Not Allowed")


class UnsupportedMediaType(BizError):
    def __init__(self):
        super().__init__(415, "Unsupported Media Type")


class InternalServerError(BizError):
    def __init__(self):
        super().__init__(500, "Internal Server Error")



from django.shortcuts import render
from django.http import JsonResponse
from django.contrib.auth.decorators import login_required
from django.urls import reverse_lazy
from django.conf import settings

from .openapi3 import get_docs
from .constants import DJANGO_APIS_OPENAPI_LOGIN_URL_KEY
from .constants import DJANGO_APIS_SWAGGER_LOGIN_REQUIRED_KEY

DJANGO_APIS_OPENAPI_LOGIN_URL = getattr(
    settings,
    DJANGO_APIS_OPENAPI_LOGIN_URL_KEY,
    reverse_lazy("admin:index"),
)
DJANGO_APIS_SWAGGER_LOGIN_REQUIRED = getattr(
    settings,
    DJANGO_APIS_SWAGGER_LOGIN_REQUIRED_KEY,
    True,
)


def swagger_ui_view(request):
    """SWAGGER-UI管理界面。"""
    return render(
        request,
        "django_apis/swagger_ui.html",
        {},
    )


def swagger_ui_data(request):
    """SWAGGER-UI数据接口。"""
    docs = get_docs()
    return JsonResponse(
        docs,
        json_dumps_params={
            "ensure_ascii": False,
        },
    )


def swagger_ui_init_js(request):
    """SWAGGER-UI JS脚本。"""
    return render(
        request,
        "django_apis/swagger_ui_init.js",
        {},
        content_type="application/javascript",
    )


if DJANGO_APIS_SWAGGER_LOGIN_REQUIRED:
    swagger_ui_view = login_required(login_url=DJANGO_APIS_OPENAPI_LOGIN_URL)(
        swagger_ui_view
    )
    swagger_ui_data = login_required(login_url=DJANGO_APIS_OPENAPI_LOGIN_URL)(
        swagger_ui_data
    )
    swagger_ui_init_js = login_required(login_url=DJANGO_APIS_OPENAPI_LOGIN_URL)(
        swagger_ui_init_js
    )



import re
import json
import inspect
import typing

from zenutils import treeutils


from pydantic import BaseModel
from pydantic import Field
from openapi_pydantic import OpenAPI
from openapi_pydantic.util import PydanticSchema
from openapi_pydantic.util import construct_open_api_with_schema_class

from django.apps import apps
from django.urls import get_resolver
from django.utils.translation import gettext
from django.conf import settings

from .constants import DJANGO_APIS_METHODS_KEY
from .constants import DJANGO_APIS_VIEW_FLAG_KEY
from .constants import DJANGO_APIS_APIVIEW_INSTANCE_KEY
from .constants import DJANGO_APIS_OPENAPI_SERVERS_KEY
from .constants import DJANGO_APIS_OPENAPI_SERVERS_DEFAULT
from .constants import DJANGO_APIS_OPENAPI_TITLE_KEY
from .constants import DJANGO_APIS_OPENAPI_VERSION_KEY
from .constants import DJANGO_APIS_OPENAPI_VERSION_DEFAULT
from .constants import DJANGO_APIS_OPENAPI_DESCRIPTION_KEY
from .constants import DJANGO_APIS_OPENAPI_DESCRIPTION_DEFAULT
from .constants import DJANGO_APIS_OPENAPI_SECURITY_DEFINITIONS_KEY
from .constants import DJANGO_APIS_OPENAPI_SCHEMAS_KEY
from .schemas import ResponseBase

DJANGO_APIS_OPENAPI_SECURITY_DEFINITIONS = getattr(
    settings,
    DJANGO_APIS_OPENAPI_SECURITY_DEFINITIONS_KEY,
    {},
)


def get_default_title():
    """获取默认的工程名称。"""
    return settings.ROOT_URLCONF.split(".")[0]


def get_tags_from_paths(paths):
    """获取所有提供了接口的应用标签列表。"""
    all_tags = []
    for path, path_info in paths.items():
        for method, method_info in path_info.items():
            tags = set(method_info.get("tags", []))
            all_tags += tags
    return set(all_tags)


def get_tags(paths):
    """获取全局应用标签信息列表。"""
    all_tags = get_tags_from_paths(paths)
    tags = []
    for app in apps.get_app_configs():
        tag_name = app.label
        if not tag_name in all_tags:
            continue
        tag = {}
        tag["name"] = tag_name
        tag["description"] = gettext(app.verbose_name)
        tags.append(tag)
    return tags


def get_query_parameters(query_schema):
    """获取url参数定义。"""
    parameters = []
    required = query_schema.get("required", [])
    for prop_name, prop_info in query_schema["properties"].items():
        prop_schema = {"type": prop_info["type"]}
        if "default" in prop_info:
            prop_schema["default"] = prop_info["default"]
        parameters.append(
            {
                "in": "query",
                "name": prop_name,
                "description": prop_info.get("description", prop_info.get("title", "")),
                "schema": prop_schema,
                "required": prop_name in required,
            }
        )
    return parameters


def has_file_field(parameter_info):
    """判断当前接口是否需要上传文件。"""
    if parameter_info.annotation == inspect._empty:
        return False
    return '"binary"' in parameter_info.annotation.schema_json()


def get_simple_type(type):
    """处理简单参数类型的定义。"""
    if issubclass(type, str):
        return {"type": "string"}
    elif issubclass(type, int):
        return {"type": "integer"}
    elif issubclass(type, float):
        return {"type": "number"}
    else:
        return {}


def get_path_parameters(parameter_name, parameter_info):
    """获取路径参数定义。"""
    return [
        {
            "in": "path",
            "name": parameter_name,
            "schema": get_simple_type(parameter_info.annotation),
            "required": True,
        }
    ]


def get_view_methods(view):
    """从视图函数中获取视图支持的HTTP请求方法列表。"""
    return getattr(
        view,
        DJANGO_APIS_METHODS_KEY,
        ["GET"],
    )


def format_tags(tags):
    """格式化标签。总是为字符串数组。"""
    if isinstance(tags, str):
        return [tags]
    elif isinstance(tags, (list, set, tuple)):
        tags = list(tags)
        tags.sort()
        return tags
    else:
        return []


def is_django_apis_view(view):
    """判断视图函数是否为使用`django-apis`提供的`apiview`方法注解的函数。"""
    return getattr(
        view,
        DJANGO_APIS_VIEW_FLAG_KEY,
        False,
    )


def get_tags_mapping():
    """创建应用标签查询表。"""
    modules_mapping = treeutils.SimpleRouterTree()
    for app in apps.get_app_configs():
        modules_mapping.index(app.module.__spec__.name, app.label)
    return modules_mapping


def make_generic_response_type(type_class, view, apiview_instance):
    """构建接口响应数据模型。"""
    name = (
        ".".join([view.__module__, view.__name__]).replace(".", "_").title()
        + "_Response"
    )
    new_type = type(
        name,
        (apiview_instance.base_response_class,),
        {
            apiview_instance.base_response_data_field: None,
            "__annotations__": {
                apiview_instance.base_response_data_field: type_class,
            },
        },
    )
    return new_type


def get_response_schema(type_class, view, apiview_instance):
    """获得接口响应数据模型。"""
    if type_class == inspect._empty:
        return {}
    elif type(type_class) == typing._GenericAlias:
        return PydanticSchema(
            schema_class=make_generic_response_type(
                type_class,
                view,
                apiview_instance,
            )
        )
    elif issubclass(type_class, ResponseBase):
        return PydanticSchema(
            schema_class=type_class,
        )
    else:
        return PydanticSchema(
            schema_class=make_generic_response_type(
                type_class,
                view,
                apiview_instance,
            )
        )


def get_securities():
    """为接口添加认证方式。"""
    securities = [{}]
    for key in DJANGO_APIS_OPENAPI_SECURITY_DEFINITIONS.keys():
        securities.append({key: []})
    return securities


def get_view_schema(view, tags=None):
    """获取视图定义。"""
    schema = {}
    description = view.__doc__ or ""
    apiview_instance = getattr(view, DJANGO_APIS_APIVIEW_INSTANCE_KEY, None)
    info = {
        "summary": description and description.splitlines()[0] or "",
        "description": description,
        "tags": format_tags(tags),
        "security": get_securities(),
    }
    signature = inspect.signature(view)
    for parameter_name, parameter_info in signature.parameters.items():
        # 处理request参数。特殊参数，不需要体现在swagger中。
        if parameter_name == "request":
            # 忽略request参数
            # 无须处理
            pass
        # 处理json payload请求体
        elif parameter_name == "payload":
            info["requestBody"] = {
                "required": True,
                "content": {
                    "application/json": {
                        "schema": (parameter_info.annotation != inspect._empty)
                        and PydanticSchema(schema_class=parameter_info.annotation)
                        or {},
                    },
                },
            }
        # 处理表单请求体
        elif parameter_name == "form":
            mimetype = "application/x-www-form-urlencoded"
            if has_file_field(parameter_info):
                mimetype = "multipart/form-data"
            info["requestBody"] = {
                "required": True,
                "content": {
                    mimetype: {
                        "schema": (parameter_info.annotation != inspect._empty)
                        and PydanticSchema(schema_class=parameter_info.annotation)
                        or {},
                    },
                },
            }
        # 处理url参数
        elif parameter_name == "query":
            if not "parameters" in info:
                info["parameters"] = []
            info["parameters"] += (
                (parameter_info.annotation != inspect._empty)
                and get_query_parameters(parameter_info.annotation.schema())
                or []
            )

        # @todo: 添加headers, cookies参数支持
        # 处理path参数
        else:
            if not "parameters" in info:
                info["parameters"] = []
            info["parameters"] += get_path_parameters(parameter_name, parameter_info)
    # 处理响应体
    info["responses"] = {
        "200": {
            "description": "Success",
            "content": {
                "application/json": {
                    "schema": get_response_schema(
                        signature.return_annotation,
                        view,
                        apiview_instance,
                    ),
                }
            },
        }
    }
    # 为每种支持的请求方法生成接口定义
    methods = get_view_methods(view)
    for method in methods:
        schema[method.lower()] = info
    return schema


def get_paths():
    """获取所有接口定义。"""
    paths = {}
    tags_mapping = get_tags_mapping()
    global_urls = get_resolver().reverse_dict
    for view_item in global_urls.lists():
        # 获取当前视图的的处理函数
        view_func = view_item[0]
        if isinstance(view_func, str):
            try:
                view_func = resolve(reverse(view_func)).func
            except Exception:
                continue
        # 如果不是django-apis视图，则忽略
        if not is_django_apis_view(view_func):
            continue
        # 查找到当前视图函数所在应用标签，作为本接口的tag_name
        tag_name, _ = tags_mapping.get_best_match(view_func.__module__)
        if not tag_name:
            tag_name = "__all__"
        # 遍历所有视图公开的paths
        view_paths = view_item[1]
        for view_path in view_paths:
            path = "/" + view_path[0][0][0]
            path = re.sub("\\%\\(([^\\)]*)\\)s", "{\\1}", path)
            paths[path] = get_view_schema(view_func, tag_name)
    return paths


class PingRequest(BaseModel):
    """Ping Request"""

    req_foo: str = Field(description="foo value of the request")
    req_bar: str = Field(description="bar value of the request")


class ErrorInfo(BaseModel):
    code: int = 0
    message: str = "OK"


class PingResponse(BaseModel):
    """Ping response"""

    resp_foo: str = Field(description="foo value of the response")
    resp_bar: str = Field(description="bar value of the response")
    error: ErrorInfo = ErrorInfo()


def get_paths2():
    return {
        "/ping": {
            "post": {
                "requestBody": {
                    "content": {
                        "application/json": {
                            "schema": PydanticSchema(schema_class=PingRequest)
                        }
                    }
                },
                "responses": {
                    "200": {
                        "description": "pong",
                        "content": {
                            "application/json": {
                                "schema": PydanticSchema(schema_class=PingResponse)
                            }
                        },
                    }
                },
            }
        }
    }


def get_docs():
    """获取swagger数据。"""
    paths = get_paths()
    tags = get_tags(paths)

    openapi = construct_open_api_with_schema_class(
        OpenAPI.model_validate(
            {
                "openapi": "3.1.0",
                "servers": getattr(
                    settings,
                    DJANGO_APIS_OPENAPI_SERVERS_KEY,
                    DJANGO_APIS_OPENAPI_SERVERS_DEFAULT,
                ),
                "info": {
                    "title": getattr(
                        settings,
                        DJANGO_APIS_OPENAPI_TITLE_KEY,
                        get_default_title(),
                    ),
                    "version": getattr(
                        settings,
                        DJANGO_APIS_OPENAPI_VERSION_KEY,
                        DJANGO_APIS_OPENAPI_VERSION_DEFAULT,
                    ),
                    "description": getattr(
                        settings,
                        DJANGO_APIS_OPENAPI_DESCRIPTION_KEY,
                        DJANGO_APIS_OPENAPI_DESCRIPTION_DEFAULT,
                    ),
                },
                "tags": tags,
                "paths": paths,
                "components": {
                    "securitySchemes": getattr(
                        settings,
                        DJANGO_APIS_OPENAPI_SECURITY_DEFINITIONS_KEY,
                        {},
                    ),
                },
            }
        )
    )
    return json.loads(
        openapi.json(
            by_alias=True,
            exclude_none=True,
        )
    )



from django.urls import path
from django.conf import settings

from . import swagger_ui
from .constants import DJANGO_APIS_SWAGGER_UI_PATH_KEY

DJANGO_APIS_SWAGGER_UI_PATH = getattr(
    settings,
    DJANGO_APIS_SWAGGER_UI_PATH_KEY,
    "docs/",
)
if DJANGO_APIS_SWAGGER_UI_PATH.startswith("/"):
    DJANGO_APIS_SWAGGER_UI_PATH = DJANGO_APIS_SWAGGER_UI_PATH[1:]
if not DJANGO_APIS_SWAGGER_UI_PATH.endswith("/"):
    DJANGO_APIS_SWAGGER_UI_PATH += "/"

urlpatterns = [
    path(
        f"{DJANGO_APIS_SWAGGER_UI_PATH}",
        swagger_ui.swagger_ui_view,
        name="django_apis_swagger_ui_view",
    ),
    path(
        f"{DJANGO_APIS_SWAGGER_UI_PATH}init.js",
        swagger_ui.swagger_ui_init_js,
        name="django_apis_swagger_ui_init_js",
    ),
    path(
        f"{DJANGO_APIS_SWAGGER_UI_PATH}data.json",
        swagger_ui.swagger_ui_data,
        name="django_apis_swagger_ui_data",
    ),
]
